﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.AI;


[RequireComponent (typeof (NavMeshAgent))]
public class EnemyController : MonoBehaviour
{

	// プレイヤーの位置情報
	[SerializeField]
	Transform			Player;

	// 巡回ルート地点をUnityEditorからセット
	[SerializeField]
	Transform[] 		PatrolRoot;

	// 歩き時の移動速度
	[SerializeField]
	float				WalkSpeed = 1f;		// 秒速

	// 走る時の移動速度
	[SerializeField]
	float				RunSpeed = 4f;		// 秒速

	// 攻撃範囲
	[SerializeField]
	float				AttackRange = 1f;	// メートル

	// 敵を発見できる距離
	[SerializeField]
	float				DistanceToFindEnemies = 5f;		// メートル

	// 敵を見失う距離
	[SerializeField]
	float				DistanceToLoseSightOfEnemies = 10f;		// メートル

	// 敵を発見できる視野
	[SerializeField]
	float				FieldOfView = 90f;		// 度



	// 敵を動かすエージェント
	NavMeshAgent	m_navMeshAgent;


	/// <summary>
	/// 敵の行動の種類.
	/// </summary>
	public enum RoutineType
	{
		Patrol,		// パトロール
		Chase,		// 追跡中
		Attack,		// 攻撃
	}
	RoutineType	m_routine = RoutineType.Patrol;


	/// <summary>
	/// 行動を変更した理由.
	/// </summary>
	enum ReasonForChange
	{
		Near,		// 近い
		Far,        // 遠い
		Damage,		// ダメージを受けた
	}
	ReasonForChange	m_reasonForChange = ReasonForChange.Far;


	// アニメーター
	Animator		m_animator;


	// Animatorのパラメータを定数として宣言
	const string kAnimatorPramWalk			= "WALK";
	const string kAnimatorPramRun			= "RUN";
	const string kAnimatorPramAttack		= "ATTACK";


	// 移動速度
	float		m_moveSpeed = 1f;		// 秒速
	int			m_patrolRootNo = -1;	// 巡回ルート管理番号 マイナス値なら強制的にルートをセット



	// Use this for initialization
	void Start () {
		// Animatorコンポーネントを取得
		m_animator = GetComponent<Animator> ();

		// NavMeshAgentコンポーネントを取得
		m_navMeshAgent = GetComponent<NavMeshAgent> ();
	}



	// Update is called once per frame
	void Update () {


		Debug.Log (m_routine);

		switch (m_routine) {
		case RoutineType.Patrol:	// 巡回
			{
				Patrol ();
			}
			break;

		case RoutineType.Chase: // 追跡
			{
				Chase ();
			}
			break;

		case RoutineType.Attack:
			{
				Attack ();
			}
			break;
		}
	}





	/// <summary>
	/// 巡回
	/// </summary>
	void Patrol()
	{
		// 経路探索の準備ができているなら
		if (m_navMeshAgent.pathStatus != NavMeshPathStatus.PathInvalid)
		{
			// プレイヤーの位置と自分の位置から角度を算出し視野に入っているかの判定を行う
			Vector3 targetDir = Player.position - transform.position;
			float angle = Vector3.Angle(targetDir, transform.forward);

			// 敵を見つけたら、
			if ((Vector3.Distance (transform.position, Player.position) < DistanceToFindEnemies) && (angle < FieldOfView))
			{
				m_reasonForChange = ReasonForChange.Near;
				m_routine = ChangeRoutine ();
			}

			// 自分の位置と目的地の距離が1m未満になったら または、初めてのルート設定の場合なら
			else if ((m_patrolRootNo < 0) || (Vector3.Distance (transform.position, PatrolRoot [m_patrolRootNo].position) < 1.0f))
			{
				// 強制セット
				if (m_patrolRootNo < 0) {
					m_patrolRootNo *= -1;
				}
				// 巡回ルート番号を進める
				else {
					// 目的地を変更(次の目的地へ)
					m_patrolRootNo++;
				}


				// 配列の数を超えたら0に戻す
				if (m_patrolRootNo >= PatrolRoot.Length)
				{
					m_patrolRootNo = 0;
				}


				// アニメーションを歩くに変更(アニメーションが切り変わらなかった場合の対策)
				m_animator.SetTrigger (kAnimatorPramWalk);
			}

			// 目的地をセット
			m_navMeshAgent.SetDestination (PatrolRoot[m_patrolRootNo].position);
		}
	}


	/// <summary>
	/// 追跡
	/// </summary>
	void Chase ()
	{

		// 経路探索の準備ができているなら
		if (m_navMeshAgent.pathStatus != NavMeshPathStatus.PathInvalid)
		{

			// プレイヤーとの距離
			float distance = Vector3.Distance (transform.position, Player.position);

			// プレイヤーとの距離が離れたなら
			if (distance >= DistanceToLoseSightOfEnemies)
			{
				// パトロールに戻す
				m_reasonForChange = ReasonForChange.Far;

				// 次のRoutineを決める
				m_routine = ChangeRoutine ();
			}
			// 攻撃が届く範囲に入ったなら
			else if (distance <= AttackRange)
			{
				// 攻撃へ
				m_reasonForChange = ReasonForChange.Near;

				// 次のRoutineを決める
				m_routine = ChangeRoutine ();
			}
			// 追跡中
			else
			{
				// 目的地をセット
				m_navMeshAgent.SetDestination (Player.position);
			}

		}
	}


	/// <summary>
	/// 攻撃処理
	/// </summary>
	void Attack()
	{
		// プレイヤーとの距離
		float distance = Vector3.Distance (transform.position, Player.position);

		// プレイヤーとの距離が離れたなら
		if (distance > AttackRange)
		{
			// パトロールに戻す
			m_reasonForChange = ReasonForChange.Far;

			// 次のRoutineを決める
			m_routine = ChangeRoutine ();
		}
		// アニメーションが終了したら もう一度攻撃
		else if ((m_animator.GetCurrentAnimatorStateInfo (0).normalizedTime > 1.0) &&
		    	(m_animator.GetCurrentAnimatorStateInfo (0).IsName ("Attack") == true))
		{
			// アニメーションを攻撃に変更
			m_animator.SetTrigger (kAnimatorPramAttack);

			// プレイヤーの方を向く
			transform.LookAt (Player);

		}
	}


	/// <summary>
	/// 敵の行動を変更.
	/// </summary>
	/// <returns>The routine.</returns>
	public RoutineType ChangeRoutine()
	{
		RoutineType ret = m_routine;

		switch (m_routine) {
		case RoutineType.Patrol:	// 巡回中に
			{
				// 敵を発見したので追跡に
				if (m_reasonForChange == ReasonForChange.Near) {
					ret = RoutineType.Chase;
				}
			}
			break;

		case RoutineType.Chase:		// 追跡中に
			{
				// 距離が離れたので巡回へ
				if (m_reasonForChange == ReasonForChange.Far) {	
					ret = RoutineType.Patrol;
				}
				// 距離が近づいたの攻撃
				else if (m_reasonForChange == ReasonForChange.Near) {
					ret = RoutineType.Attack;
				}
			}
			break;

		case RoutineType.Attack:	// 攻撃中に
			{
				// 距離が離れたので追跡へ
				if (m_reasonForChange == ReasonForChange.Far) {	
					ret = RoutineType.Chase;
				}
			}
			break;

		}


		// 次の行動を開始
		StartRoutine (ret);

		return ret;
	}



	/// <summary>
	/// Routineを変えたときの初期化処理.
	/// </summary>
	/// <param name="value">変更後のRoutine.</param>
	void StartRoutine(RoutineType value)
	{

		switch (value) {
		case RoutineType.Patrol:	// 巡回
			{
				// アニメーションを歩くに変更
				m_animator.SetTrigger (kAnimatorPramWalk);

				// 移動速度をセット
				m_moveSpeed = WalkSpeed;

				// 一番近い巡回目的地を探す
				float min = float.MaxValue;	// float型の最大値　最小距離を記録する用
				for (int i = 0; i < PatrolRoot.Length; i++)
				{
					// 現在位置と経由地の距離を調べる
					float distance = Vector3.Distance (transform.position, PatrolRoot [i].position);

					// 求めた距離が今までの最小距離より小さいなら
					if (distance < min)
					{
						min = distance;			// 最小距離を更新
						m_patrolRootNo = -i;	// 最小距離を出した番号を利用(強制的にセットするためにマイナスの値に)
					}
				}
			}
			break;

		case RoutineType.Chase: // 追跡
			{
				// アニメーションを走るに変更
				m_animator.SetTrigger (kAnimatorPramRun);

				// 移動速度をセット
				m_moveSpeed = RunSpeed;
			}
			break;


		case RoutineType.Attack: // 攻撃
			{
				// アニメーションを攻撃に変更
				m_animator.SetTrigger (kAnimatorPramAttack);

				// プレイヤーの方を向く
				transform.LookAt (Player);
			}
			break;


		}

		// 速度を反映
		m_navMeshAgent.speed = m_moveSpeed;
	}


}